"""
Created on Fri Apr 26 16:39:50 2019

@author: chris.kerklaan

Vector wrapper used for GIS programming

Goals:
    1. Provide a better interface for ogr
    2. Provides a table and spatial index which grows and deletes with adding
    and deleting features
    3. Provides basic geoprocessing options: Clipping, dissolving etc.
    4. Provides appropriate fixing of geometries

TODO for rewriting in shapely tre version 1.0:
    1. Make shapely available in tre's geometry functions.
        1.1. Add svg in geometries. #Done
        1.2. Convert easily between shapely and ogr. #Done
    2. Rewrite functions below vector.py so that they use the Vector class.
        2.0. Rewrite vectorbase so that it is as fast as possible.
        2.1. Check which functions are used within other functions.
        2.2 Voronoi can be rewritten with the multipoint voronoi.
        
    3. Rewrite the table and table features with numpy.
    4. Add @cached_property
    5. Accelerate table generation.
    6. We can remove rtree indexing as it is already present within shapely/pygeos.
    6. Finally do the testing.
    
Bugs:
    1. Multipart2singlepart does not write as single part file shape

Note:
    21. Ogr indices are created when copying the vector

Notes on adding shapely:
    There are multiple reasons for adding shapely:
        - Speed-ups for multi-geometry intersections.
        - Extra geometry operations.
    We still need to have a fast way to use the shapely functions.
    
        

Notes on pointers:
    Datasource is your pointer, pointer always needs to be present when
    data is edited in layer
    When a feature is created the feature is the pointer.

    Due to this pointers, gdal ds, layers and features are kept in place,
    hence the classs are not an extension of the object but built on top of them.
    If working with pure ogr, make sure that pointer are available:

        lines = Vector(TEST_DIRECTORY+"lines.shp")
        works:
            layer = lines.layer
            feature=layer[0]
            feature.geometry().Clone()
        does not work:
            layer = lines.layer
            feature.geometry().Clone()

    Also:
        works:
            for geometry in multipartgeometry:
                geometry.GetArea()
        does not work:
            for geometry in multipartgeometry:
                function(geometry)

            def function(geometry):
                geometry.GetArea()

    Functions outside the class can only use ogr-functions
    
Notes on gdal speed-ups
    - Function with 'Get' are the fastest gdal functions
        (GetFeature, GetGeometry) instead of indexing '[]' and .geometry() and .items()
        Use field_indices and feature_items to bypass and increase speed

"""
# First-party imports
import logging
import importlib
import numpy as np

# from multiprocessing import Pool

# Third party imports
from osgeo import ogr, gdal

HAS_SCIPY = importlib.util.find_spec("scipy") is not None
if HAS_SCIPY:
    from scipy import spatial

# Local imports

from threedi_edits.gis.geometry import (
    SINGLE_TYPES,
    SINGLE_TO_MULTIPLE,
    MULTI_TYPES,
    clip as clip_geometry,
    difference as difference_geometry,
    dissolve as dissolve_geometry,
)

from threedi_edits.gis.polygon import centerline as centerline_geometry
from threedi_edits.gis.polygon import Polygon
from threedi_edits.gis.vectorbase import VectorBase, mem_path
from threedi_edits.utils.project import Classes, Functions
from threedi_edits.utils.progress import Progress

# structure
classes = Classes(__name__, local_only=True)
functions = Functions(__name__)


# Global DRIVERS
DRIVER_GDAL_MEM = gdal.GetDriverByName("MEM")
DRIVER_OGR_MEM = ogr.GetDriverByName("Memory")
DRIVER_OGR_GPKG = ogr.GetDriverByName("GPKG")
DRIVER_OGR_SHAPEFILE = ogr.GetDriverByName("ESRI Shapefile")
DRIVERS_SQL = ["PostgreSQL", "GPKG"]

# Memory counter
_mem_num = 0

# Logger
logger = logging.getLogger(__name__)


class Vector(VectorBase):
    """wrapper of ogr layer, input is ogr datasource, only one layer
    is expected
    """

    instance = "gis.vector.Vector"

    def __init__(self, path, **kwargs):
        super().__init__(path, **kwargs)

    def copy(self, geometry_type=None, shell=False, fields=True, driver="GPKG"):
        """
        Copies a vector.
        params:
            geometry_type: change the geometry type
            shell: Does not add the features
            fields: Set to True is fields have to be included
        """
        return Vector(
            path=None, datasource=self._copy(geometry_type, shell, fields, driver)[0]
        )

    def subset(self, fids):
        """Returns a subset of features based on a list of fids"""
        return Vector(self._subset(*fids))

    def spatial_subset(
        self,
        geometry: ogr.Geometry,
        method: str = "intersects",
    ):
        """
        Spatially filtered geometries using the shapely predicates.
        params:
            geometry: ogr.Geometry, shapely geometry, te.Geometry
            method: Use the shapely predicates.

        """
        if hasattr(geometry, "ds"):
            return self.full_spatial_filter(geometry, method)

        return Vector(self._subset(*self._sfilter(geometry, method)))

    def filter(self, **filtering):
        """
        Filters the dataset on fields and values of those fields.
        E.g., percentile=1 or class=1 or both.
        """
        return Vector(self._filter(**filtering))

    def spatial_filter(
        self,
        geometry: ogr.Geometry,
        method: str = "intersects",
    ):
        """
        Spatially filtered geometries using the shapely predicates.
        params:
            geometry: ogr.Geometry, shapely geometry, te.Geometry
            method: Use the shapely predicates.
        """
        print("deprecated, use spatial_subset")

        if hasattr(geometry, "ds"):
            return self.full_spatial_filter(geometry, method)

        return Vector(self._subset(*self._sfilter(geometry, method)))

    def add(self, feature=None, geometry=None, fid=-1, **items):
        """adds a feature"""
        return self._add(feature, geometry, fid, **items)

    def full_spatial_subset(self, vector, method="intersects", quiet=True):
        """
        Spatially filteres with a whole vector.
        """
        fids = []
        for geometry in Progress(vector.geometries, f"{method} filter", quiet):
            fids.extend(self._sfilter(geometry, method))

        return Vector(self._subset(*fids))

    def buffer(self, buffer_size: float, quiet: bool = True):
        """Buffers a layer, output is always a polygon"""
        target = self.copy(geometry_type=ogr.wkbPolygon, shell=True)
        for feature in Progress(self, "Buffer", quiet):
            buffered = feature.geometry.Buffer(buffer_size)
            target.add(geometry=buffered, items=feature.items)
        return target

    def simplify(self, simple_factor: float, quiet: bool = True):
        """Simplifies all geometries."""
        target = self.copy(geometry_type=ogr.wkbPolygon, shell=True)
        for feature in Progress(self, "Simplify", quiet):
            simple = feature.geometry.Simplify(simple_factor)
            target.add(geometry=simple, items=feature.items)
        return target

    def centroid(self, quiet: bool = True):
        """Creates a centroid of all geometries."""
        target = self.copy(geometry_type=ogr.wkbPoint, shell=True)
        for feature in Progress(self, "Centroid", quiet):
            centroid = feature.geometry.Centroid()
            target.add(geometry=centroid, items=feature.items)
        return target

    def to_single(self, quiet: bool = True):
        """Converts multiparts to single parts"""
        target = self.copy(geometry_type=SINGLE_TYPES[self.geometry_type], shell=True)
        for feature in Progress(self, "To singles", quiet):
            geometry_type = feature.geometry.type
            items = feature.items
            if geometry_type > 3:
                for part in feature.geometry:
                    target.add(items=items, geometry=part)
            else:
                target.add(items=items, geometry=feature.geometry)
        return target

    def reproject(self, epsg: int, quiet=True):
        """Takes an ogr layer, reprojects and returns an ogr datasource"""
        target = Vector.from_scratch("reprojected", self.geometry_type, epsg)
        for field_name, field_type in self.fields.items.items():
            target.add_field(field_name, field_type[0])

        for feature in Progress(self, "Reproject", quiet):
            geometry = feature.geometry.reproject(epsg)
            target.add(geometry=geometry, items=feature.items)

        return target

    def centerlines(self, resolution, quiet: bool = True):
        """creates centerlines from polygons"""
        target = self.copy(geometry_type=ogr.wkbLineString, shell=True)
        for feature in Progress(self, "Centerline", quiet):
            centerline = centerline_geometry(feature.geometry, resolution)
            target.add(geometry=centerline, items=feature.items)
        return target

    def polygon_to_lines(self, quiet=True):
        """Takes a polygon and returns a linestring"""
        target = self.copy(geometry_type=ogr.wkbLineString, shell=True)
        for feature in Progress(self, "Centerline", quiet):
            target.add(geometry=feature.geometry.Boundary(), items=feature.items)
        return target

    def fix(self, geometry_type=None, quiet=True):
        """
        Fixes the geometries of a layer.
        Invalid geometries but also 3D features will be
        set to the layers' geometry type.
        params:
            geometry_type: Sets the target geometry type
                1 = point
                2 = linestring
                3 = polygon
                4 = multipoint
                5 = multilinestring
                6 = multipolygon
            quiet: Shows the progress

        """
        if not geometry_type:
            geometry_type = self.geometry_type

        target = self.copy(geometry_type=geometry_type, shell=True)
        for feature in Progress(self, "geometry fixes", quiet):
            target.add(
                geometry=feature.geometry.fix(geometry_type), items=feature.items
            )
        return target

    def clip(
        self,
        mask=None,
        geometry: ogr.Geometry = None,
        quiet: bool = True,
    ):
        """
        Clip off areas that are not covered by the input vector.
        If the feature falls outside of the masks, the will not be included.
        We are skipping features without geometries.


        params:
            mask: Vector on which it is clipped.
            quiet: Shows the progress
            use_ogr: uses the ogr variant of this function
        """
        if geometry:
            mask = Vector.from_scratch("", 3, geometry.epsg)
            mask.add(geometry=geometry)

        target = self.copy(shell=True)
        type_geometry = target.geometry_type
        for feature in Progress(self, "Clipping", quiet):
            geometry = feature.geometry
            if geometry is None:
                continue
            for mask_feature in mask.spatial_subset(geometry):
                clipped = clip_geometry(geometry, mask_feature.geometry, type_geometry)
                if clipped is None:
                    continue
                for clip in clipped:
                    target.add(geometry=clip, items=feature.items)

        return target

    def difference(self, mask, quiet: bool = True, ignore_errors: bool = False):
        """
        This function takes a difference between vector layer and difference layer.
        - Takes into account multiparts and single parts.
        - It also leaves geometries which are not valid.
        params:
            mask: Vector in which the difference is taken.
            quiet: Shows progress or not.
            ignore_errors: Ignores errors if True.

        """

        geometry_types = SINGLE_TO_MULTIPLE[self.geometry_type]
        target = self.copy(shell=True)
        for feature in Progress(self, "Difference", quiet):
            if feature is None:
                continue

            difference = feature.geometry
            masks = mask.spatial_subset(difference)

            for loop, mask_feature in enumerate(masks):
                mask_geometry = mask_feature.geometry
                difference = difference_geometry(
                    difference, mask_geometry, ignore_errors
                )

                # print if multiplication of 1000 features
                if loop % 5000 == 0 and loop > 0:
                    print(f"Doing difference at {loop} mask features")

            diff_part_type = difference.GetGeometryType()
            if diff_part_type not in geometry_types:
                if diff_part_type == ogr.wkbGeometryCollection:
                    for part in difference:
                        if part.GetGeometryType() == self.geometry_type:
                            target.add(geometry=part, items=feature.items)
            else:
                target.add(geometry=difference, items=feature.items)
        return target

    def spatial_clusters(self, quiet=True) -> (list, list):
        """
        Returns a list of ids of geometries that intersect each other.
        Returns a list of multipolygons of clustered geometries.
        """
        # TODO does not work yet.
        work = self.copy(shell=False)
        target = self.copy(shell=True)
        target.add_field("dids", str)

        total = len(work)
        work.reset()

        progress = Progress(message="Spatial clustering", total=total, quiet=quiet)

        # keep repeating until the work layer is empty
        while len(work) != 0:

            # new cluster id list
            cluster_ids = []
            cluster_geometry = ogr.Geometry(ogr.wkbMultiPolygon)

            # get first feature, geometry, first spatial filter, delete from work
            feature = next(work)
            if feature is None:
                continue

            cluster_ids.append(feature.id)
            work.delete(feature.id)
            cluster_geometry.AddGeometry(feature.geometry)
            cluster_geometry = ogr.ForceToMultiPolygon(cluster_geometry.UnionCascaded())

            # keep repeating until no more growth of the cluster
            while True:
                growth = False  # we are not growing
                for intersect in work.spatial_subset(cluster_geometry):
                    intersect_geometry = intersect.geometry
                    # we are growing
                    growth = True
                    # add this to the cluster
                    cluster_ids.append(intersect.id)
                    cluster_geometry.AddGeometry(intersect_geometry)
                    cluster_geometry = ogr.ForceToMultiPolygon(
                        cluster_geometry.UnionCascaded()
                    )

                    work.delete(intersect.id)

                if not growth:  # no more growth
                    break

            # add the cluster to the cluster_list

            dids = "_".join([str(c) for c in cluster_ids])
            target.add(geometry=cluster_geometry, dids=dids)
            progress.show(custom_add=len(cluster_ids))

        return target

    def dissolve(self, field: str = None, quiet: bool = True):
        """
        Dissolves the vector into a single multipolygon feature.
        params:
            field: The vector will be dissolved on the fields' unique values.
            quiet: verbosity.
        """

        multi_type = MULTI_TYPES[self.geometry_type]
        target = self.copy(shell=True, geometry_type=multi_type)
        if field is not None:  # finds unique fields, then dissolves

            unique = {f: [] for f in set(self.table[field])}

            for feature in self:
                unique[feature[field]].append(feature.id)

            for field_value, fid_list in Progress(
                unique.items(), f"Dissolve field {field}", quiet
            ):
                multi = ogr.Geometry(multi_type)
                for fid in fid_list:
                    multi.AddGeometry(self[fid].geometry)

                target.add(geometry=dissolve_geometry(multi))

        else:
            # here we are just dissolving all
            singles = self.to_single()
            multi = ogr.Geometry(multi_type)
            for feature in singles:
                multi.AddGeometry(feature.geometry)

            logger.info("Dissolving all")
            target.add(geometry=dissolve_geometry(multi))

        return target

    def voronoi(self, buffer=1, quiet=True):
        """
        Create vonoroi polygons based on a point layer.
        params:
            Buffer: used to include outer points.
            quiet: verbosity.

        """
        target = self.copy(shell=True, geometry_type=ogr.wkbPolygon)
        extent_geometry = self.extent_geometry

        point_list = []
        point_items = {}
        for fid, feature in Progress(
            self, "Vonoroi - Setting input points", enum=True, quiet=quiet
        ):
            point_list.append([*feature.geometry.points[0]])
            point_items[fid] = feature.items

        # add extent to properly add edge points
        for line in extent_geometry:
            for point in line.GetPoints():
                point_list.append([point[0], point[1]])

        vor = spatial.Voronoi(point_list)
        for fid, region_index in Progress(
            vor.point_region,
            "Vonoroi - Processing output polygons",
            enum=True,
            quiet=quiet,
        ):
            rpoints = []
            for vertex_index in vor.regions[region_index]:
                if vertex_index != -1:  # the library uses this for infinity
                    rpoints.append(list(vor.vertices[vertex_index]))
            rpoints.append(rpoints[0])
            if len(rpoints) < 4 or fid > len(point_items) - 1:  # anders geen polygon
                continue

            polygon = Polygon.from_points(rpoints)
            if not polygon.Within(extent_geometry):
                polygon = polygon.Intersection(extent_geometry)

            target.add(geometry=polygon, items=point_items[fid])
        return target

    def rasterize(
        self,
        rows=None,
        columns=None,
        geotransform=None,
        resolution=None,
        nodata=-9999,
        field=None,
        extent=None,
        all_touches=False,
        options=None,
        return_ds=True,
        data_type=gdal.GDT_Float32,
    ):
        """Rasterizes the vector as a boolean,
        If field or all touches is given that is used
        """

        if not columns and not rows:
            if not resolution:
                logger.error("please provide resolution")
                return
            if not extent:
                extent = self.extent
            x1, x2, y1, y2 = extent
            columns = int((x2 - x1) / resolution)
            rows = int((y2 - y1) / resolution)
            geotransform = (x1, resolution, 0, y2, 0, -resolution)

        if data_type == float:
            data_type = gdal.GDT_Float32
        elif data_type == int:
            data_type = gdal.GDT_Int32

        return rasterize(
            self.layer,
            rows,
            columns,
            geotransform,
            self.spatial_reference.wkt,
            nodata,
            field,
            all_touches,
            options,
            return_ds,
            data_type,
        )

    def interpolate(
        self,
        field,
        resolution,
        algorithm="invdist",
        nodata_value=-9999,
        quiet=True,
        **arguments,
    ):
        """
        Returns a gdal datasource.  Can be loaded with Raster.

        Interpolates vector points using gdal grid
        https://gdal.org/programs/gdal_grid.html#interpolation-algorithms


        params:
            field: Fields of which is used to interpolate
            resolution: Resulution of the final raster
            algorithm: See the above link to see how it can be filled.
            argument: See the above link to how it can be filled.

        """
        logger.debug("vector - interpolate - Start")
        progress = Progress(total=100, message="GDALGrid Interpolate", quiet=quiet)
        argument = ""
        for k, v in arguments.items():
            argument = argument + f"{k}={v}:"

        if len(argument) == 0:
            raise ValueError(
                "Please fill in an arguments, look at (https://gdal.org/programs/gdal_grid.html#interpolation-algorithms)"
            )

        assert field in self.fields

        extent = self.extent
        width = int((extent[1] - extent[0]) / resolution)
        height = int((extent[3] - extent[2]) / resolution)
        grid_options = gdal.GridOptions(
            zfield=field,
            algorithm=f"{algorithm}:{argument}",
            width=width,
            height=height,
            callback=progress.gdal,
        )

        path = mem_path()
        ds = gdal.Grid(
            destName=path,
            srcDS=self.as_gdal_dataset,
            options=grid_options,
        )

        # The raster which is extracted from the gdalGrid is acutally upside
        # down =with epsg 28992.
        # So we have to flip the array and change the geotransform.

        # We will now reverse it to the left upper corner.
        transform = list(ds.GetGeoTransform())
        transform[3] = transform[3] + (height * resolution)
        transform[-1] = -transform[-1]
        ds.SetGeoTransform(tuple(transform))

        band = ds.GetRasterBand(1)
        array = band.ReadAsArray()
        band.WriteArray(np.flip(array, 0), 0, 0)

        # set nodata value
        band.SetNoDataValue(-9999)

        band = None
        ds = None

        return gdal.Open(path)


def rasterize(
    layer,
    rows,
    columns,
    geotransform,
    spatial_reference_wkt,
    nodata=-9999,
    field=None,
    all_touches=False,
    options=None,
    return_ds=False,
    data_type=gdal.GDT_Float32,
):
    target_ds = DRIVER_GDAL_MEM.Create("", columns, rows, 1, data_type)

    # set nodata
    band = target_ds.GetRasterBand(1)
    band.SetNoDataValue(nodata)
    band.Fill(nodata)
    band.FlushCache()

    # set metadata
    target_ds.SetProjection(spatial_reference_wkt)
    target_ds.SetGeoTransform(geotransform)

    # set options
    gdal_options = []

    if field:
        logger.info(field)
        gdal_options.append(f"ATTRIBUTE={field}")

    if all_touches:
        gdal_options.append("ALL_TOUCHES=TRUE")

    if options:
        gdal_options.extend(options)

    if len(gdal_options) == 0:
        gdal.RasterizeLayer(target_ds, (1,), layer, burn_values=(1,))
    else:
        gdal.RasterizeLayer(target_ds, [1], layer, options=gdal_options)

    if return_ds:
        return target_ds
    else:
        array = target_ds.ReadAsArray()
        target_ds = None
        return array



# -*- coding: utf-8 -*-
"""
Created on Fri Feb 12 16:45:07 2021

@author: chris.kerklaan
"""

# First-party imports
import os
import pathlib

# Third-party imports
from osgeo import ogr

# Local imports
from .vector import Vector
from .utils import DRIVERS
from threedi_edits.utils.project import Classes, Functions

# structure
classes = Classes(__name__, local_only=True)
functions = Functions(__name__)


# globals
OGR_MEM_DRIVER = ogr.GetDriverByName("Memory")
OPTIONS = ["OVERWRITE=YES", "FID=fid"]


class VectorGroup:
    instance = "gis.vector.VectorGroup"

    def __init__(
        self,
        path=None,
        ogr_ds=None,
        options={},
        memory=True,
        write=False,
    ):
        """vector groups"""
        if path is not None:
            self.ext = pathlib.Path(str(path)).suffix
            self.ds = ogr.Open(str(path), write, **options)

        elif ogr_ds is not None:
            self.ds = ogr_ds

        if memory:
            self.ds = self._copy()

    @classmethod
    def from_pg(cls, host, port, user, password, dbname, write=False, memory=False):
        pg_string = ("PG:host={} port={} user='{}'" "password='{}' dbname='{}'").format(
            host, port, user, password, dbname
        )

        return cls(ogr_ds=ogr.Open(pg_string, write), memory=memory)

    @classmethod
    def from_scratch(cls, name="scratch"):
        return cls(ogr_ds=OGR_MEM_DRIVER.CreateDataSource(name), memory=False)

    def __getitem__(self, i):
        return Vector.from_ds(self.ds, layer_name=i)

    def __iter__(self):
        for layer in self.layers:
            yield self[layer]

    @property
    def layers(self):
        return [layer.GetName() for layer in self.ds]

    @property
    def drivers(self):
        return DRIVERS

    def add(self, vector, name):
        vector.layer_name = name
        self.ds.CopyLayer(vector.layer, name, options=["OVERWRITE=YES"])

    def add_styling(self, vector, name="layer_styles"):
        """add styling to the gpkg by adding the layer_styles column"""
        self.add(vector, name)

    def clear(self):
        for table in self:
            table.delete_all()

    def _copy(self):
        return OGR_MEM_DRIVER.CopyDataSource(self.ds, "")

    def write(self, path, overwrite=False):
        path = str(path)  # pathlib

        driver_name = self.drivers[os.path.splitext(path)[-1]]
        driver = ogr.GetDriverByName(driver_name)

        if overwrite:
            if os.path.exists(path):
                driver.DeleteDataSource(path)

        # create output file
        out_ds = driver.CreateDataSource(path)

        for layer in self:
            if "fid" in layer.fields:
                layer.delete_field("fid")
            out_ds.CopyLayer(layer.layer, layer.name, options=OPTIONS)

        out_ds = None

    def execute_sql(self, sql, dialect=None):
        self.ds.ExecuteSQL(sql, dialect)

    def clip(self, vector):
        output = VectorGroup.from_scratch()
        for layer in self:
            output.add(layer.clip(vector), layer.name)
        return output

    def __repr__(self):
        return str(self.layers)



# -*- coding: utf-8 -*-
"""
Created on Mon Feb 15 09:04:00 2021

@author: chris.kerklaan
"""

_IDK = "¯\_(ツ)_/¯"
_ANGRY = "(╯°□°）╯︵ ┻━┻ ┻━┻ ︵ ヽ(°□°ヽ)"
_HAPPY = "＼(^o^)／"


# warnings and errors
def custom_formatwarning(msg, *args, **kwargs):
    # ignore everything except the message
    return f"{_IDK} (Warning):" + str(msg) + "\n"


class ThreediValueTypeError(Exception):
    def __init__(self, message, table, key, value):
        print(_ANGRY)
        self.message = message
        self.table = table
        self.key = key
        self.value = value


class ThreediConnectedTableError(Exception):
    def __init__(self, message, table, connected_table, field, connection_id):
        print(_ANGRY)
        self.message = message
        self.table = table
        self.connected_table = connected_table
        self.connection_id = connection_id
        self.field = field


class ThreediValueTypeWarning(UserWarning):
    def __init__(self, message, table, key, value):
        self.message = message
        self.table = table
        self.key = key
        self.value = value



# -*- coding: utf-8 -*-
"""
Created on Tue Apr 30 11:56:20 2019

@author: chris.kerklaan - N&S
"""
# First-party imports
import os
import sys
import json
import logging
import datetime
import shutil
import tempfile
from pathlib import Path

from inspect import getmembers, isfunction, isclass, ismodule

# Globals
time_last_print_small = datetime.datetime.now()
time_last_print_large = datetime.datetime.now()


def log_time(log_type, s1="", s2="", size="s"):
    print_time(s1=s1, s2=s2, size=size)

    if log_type == "info":
        logging.info("%s - %s", s1, s2)
    elif log_type == "debug":
        logging.debug("%s - %s", s1, s2)
    elif log_type == "warning":
        logging.warning("%s - %s", s1, s2)
    elif log_type == "error":
        logging.error("%s - %s", s1, s2)
    else:
        print("no log")


class Logger(object):
    def __init__(self, path, quiet=False):
        # self.terminal = sys.stdout
        self.log = open(path, "a")
        self.quiet = quiet
        self.path = path

    def set_subject(self, subject):
        self.s1 = subject

    def show(self, s2="", quiet=None, size="s"):
        # if self.quiet or quiet:
        #     be_quiet=True
        # else:
        #     be_quiet=False

        # if not be_quiet:
        global time_last_print_small
        global time_last_print_large

        now = datetime.datetime.now()
        if size == "s":
            _print = "{} - {} - {} - {}".format(
                now, now - time_last_print_small, self.s1, s2
            )

        elif size == "l":
            _print = "{} - {} - {} - {}".format(
                now, now - time_last_print_large, self.s1, s2
            )

        else:
            pass

        time_last_print_small = now
        time_last_print_large = now

        print("\n" + _print + "\n")
        self.log.write(_print + "\n")

    def close(self):
        self.log = None

    def write(self):
        self.log = None
        self.log = open(self.path, "a")


def print_time(s1="", s2="", size="s", quiet=False):
    if not quiet:
        global time_last_print_small
        global time_last_print_large

        now = datetime.datetime.now()
        if size == "s":
            _print = "{} - {} - {} - {}".format(
                now, now - time_last_print_small, s1, s2
            )

        elif size == "l":
            _print = "{} - {} - {} - {}".format(
                now, now - time_last_print_large, s1, s2
            )

        else:
            pass

        time_last_print_small = now
        time_last_print_large = now
        return print(_print + "\n")


def percentage(count, total):
    return str((count / total) * 100) + "%"


def mk_temp(path=os.getcwd()):
    tempfolder = os.path.join(path, "temp")
    if not os.path.exists(tempfolder):
        os.mkdir(tempfolder)
    return tempfolder


def print_list(_list, subject):
    if not isinstance(_list, list):
        _list = [_list]

    print("\n {}:".format(subject))
    for path in _list:
        print("\t{}".format(str(path)))


def print_dictionary(_dict, subject):
    print("\n {}:".format(subject))
    for key, value in _dict.items():
        print("\t{}:\t\t{}".format(str(key), str(value)))


def write_dictionary(path, _dict):
    with open(path, "w") as write_file:
        json.dump(_dict, write_file, indent=4)


def load_dictionary(path):
    with open(path) as json_file:
        data = json.load(json_file)
    return data


def mk_dir(path=os.getcwd(), folder_name="temp", overwrite=True):
    tempfolder = os.path.join(path, folder_name)

    if os.path.exists(tempfolder) and overwrite:
        shutil.rmtree(tempfolder)
        os.mkdir(tempfolder)
    elif not os.path.exists(tempfolder):
        os.mkdir(tempfolder)
    else:
        pass  # Folder exits, path
    return tempfolder


def create_temporary_copy(path):
    temp = tempfile.NamedTemporaryFile("w+t", delete=True)
    shutil.copy2(temp.name, path)
    return temp.name


class Progress:
    def __init__(self, total, message, size=30, file=sys.stdout):
        self.total = total
        self.message = message
        self.size = size
        self.file = file
        self.i = 0
        self.time = datetime.datetime.now()

        # clear the current line
        print("\r")

    def update(self, quiet=False):
        if quiet:
            return

        self.i += 1
        x = int(self.size * self.i / self.total)

        print(
            f"\r{self.message}: [{'.'*x}{' '*(self.size-x)}] {self.i}/{self.total}",
            end="\r",
            flush=True,
        )

        if self.i >= self.total:
            end_time = datetime.datetime.now() - self.time
            print(
                f"\r{self.message}: Completed {self.total} loops in {end_time}",
                end="\r",
                flush=True,
            )


def pretty_dict(d, indent=0):
    string = ""
    for key, value in d.items():
        string = string + "\t" * indent + str(key)
        if isinstance(value, dict):
            pretty_dict(value, indent + 1)
        else:
            string = string + "\t" * (indent + 1) + str(value)
    return string


class DisplayablePath(object):
    display_filename_prefix_middle = "├──"
    display_filename_prefix_last = "└──"
    display_parent_prefix_middle = "    "
    display_parent_prefix_last = "│   "

    def __init__(self, path, parent_path, is_last):
        self.path = Path(str(path))
        self.parent = parent_path
        self.is_last = is_last
        if self.parent:
            self.depth = self.parent.depth + 1
        else:
            self.depth = 0

    @property
    def displayname(self):
        if self.path.is_dir():
            return self.path.name + "/"
        return self.path.name

    @classmethod
    def make_tree(cls, root, parent=None, is_last=False, criteria=None):
        root = Path(str(root))
        criteria = criteria or cls._default_criteria

        displayable_root = cls(root, parent, is_last)
        yield displayable_root

        children = sorted(
            list(path for path in root.iterdir() if criteria(path)),
            key=lambda s: str(s).lower(),
        )
        count = 1
        for path in children:
            is_last = count == len(children)
            if path.is_dir():
                yield from cls.make_tree(
                    path, parent=displayable_root, is_last=is_last, criteria=criteria
                )
            else:
                yield cls(path, displayable_root, is_last)
            count += 1

    @classmethod
    def _default_criteria(cls, path):
        return True

    def displayable(self):
        if self.parent is None:
            return self.displayname

        _filename_prefix = (
            self.display_filename_prefix_last
            if self.is_last
            else self.display_filename_prefix_middle
        )

        parts = ["{!s} {!s}".format(_filename_prefix, self.displayname)]

        parent = self.parent
        while parent and parent.parent is not None:
            parts.append(
                self.display_parent_prefix_middle
                if parent.is_last
                else self.display_parent_prefix_last
            )
            parent = parent.parent

        return "".join(reversed(parts))


class Files:
    def __init__(
        self,
        file=None,
        root=None,
        folder_only=False,
        folder_filters=["__pycache__", "data"],
        file_filters=[".pyc"],
    ):

        if file:
            root = Path(file).parents[0]

        self.paths = DisplayablePath.make_tree(root)
        self.filter(folder_only, file_filters, folder_filters)

    def __repr__(self):
        return self.general

    def filter(self, folder_only, files=[], folders=[]):

        self.general = ""
        for path in self.paths:
            path_display = path.displayable()

            # flters on folders only
            if folder_only:
                if "." in path_display:
                    continue

            # filters on files
            do_continue = False
            for file in files:
                if path_display.endswith(file):
                    do_continue = True

            if do_continue:
                continue

            do_continue = False

            parents = []
            parent = path.parent

            # if one of the upper parents is that folder, it is skipped
            while parent != None:
                if parent:
                    parents.append(parent.displayname)

                parent = parent.parent

            do_continue = False
            for folder in folders:
                do_continue = len([p for p in parents if p == folder + "/"]) > 0

                if do_continue:
                    break

            for folder in folders:
                if folder + "/" == path.displayname:
                    do_continue = True
                    break

            if do_continue:
                continue

            self.general = self.general + path.displayable() + "\n"


class Functions:
    """retrieves all functions of a script"""

    def __init__(self, module_name):
        self.module = sys.modules[module_name]

    def __repr__(self):
        members = "Functions: \n"
        for name, a in getmembers(self.module, isfunction):
            members = members + "\t-\t" + name + "\n"

        return members


class Classes:
    """retrieves all classes"""

    def __init__(self, module_name, local_only=False):
        self.module_name = module_name
        self.module = sys.modules[module_name]
        self.local_only = local_only

    def __repr__(self):
        members = "Classes: \n"
        for name, _cls in getmembers(self.module, isclass):
            if self.local_only:
                if _cls.__module__ == self.module_name:
                    members = members + "\t-\t" + name + "\n"
            else:
                members = members + "\t-\t" + name + "\n"

        return members


class Modules:
    """retrieves all classes"""

    def __init__(self, module_name):
        self.module = sys.modules[module_name]

    def __repr__(self):
        members = "Modules: \n"
        for name, a in getmembers(self.module, ismodule):
            if not name.isupper():
                members = members + "\t-\t" + name + "\n"

        return members



from setuptools import setup, find_packages


__version__ = "1.1"

long_description = "\n\n".join([open("README.rst").read(), open("CHANGES.rst").read()])

install_requires = ["scipy", "cached_property", "shapely==2.0a1", "threedigrid_builder"]

tests_require = [
    "pytest",
    "mock",
    "pytest-cov",
    "pytest-flakes",
    "pytest-black",
    "pytest-benchmark",
]

setup(
    name="threedi-edits",
    version=__version__,
    description="An experimental pythonic 3Di schematisation api. Using this api, we can access, alter en write a 3Di database within python. Within the package gis tools are provided as well. ",
    long_description=long_description,
    long_description_content_type='text/x-rst',
    # Get strings from http://www.python.org/pypi?%3Aaction=list_classifiers
    classifiers=["Programming Language :: Python", "Framework :: Django"],
    keywords=["3Di", "GDAL", "api"],
    author="Chris Kerklaan",
    author_email="chris-kerklaan@live.nl",
    url="https://github.com/threedi-edits",
    license="LGPL",
    packages=find_packages(),
    include_package_data=True,
    zip_safe=False,
    install_requires=install_requires,
    tests_require=tests_require,
    python_requires="<3.11",
    extras_require={"test": tests_require},
    entry_points={
        "console_scripts": ["run-threedi-edits = threedi_edits.scripts:main"]
    },
)



# -*- coding: utf-8 -*-
"""
Created on Mon Dec 20 09:14:30 2021

@author: chris.kerklaan
"""
import os
import numpy as np
from threedi_edits import Vector, Raster, Progress

os.chdir(r"C:\Users\chris.kerklaan\Documents\Projecten\basis kaarten hhnk\processing")
levee_path = "levees_19gz2_types_2.shp"
dem_path = r"\\utr-3fs-01.nens.local\WorkDir\J_vLieshout\HHNK_AHN_basiskaarten\AHN4_Interpolated\ahn4_interpolated\19gz2.tif"
levees = Vector(levee_path)
dem = Raster(dem_path)
output = levees.copy(shell=True)
output.add_field("height", float)

output_merged = levees.copy(shell=True)
output_merged.add_field("height", float)
for levee in Progress(levees, "Sampling levees"):
    lines = levee.geometry.perpendicular_lines(distance=2, perpendicular_length=5)

    heights = []
    for line in lines:
        try:
            height = dem.read(line)
        except Exception:
            height = 0

        output.add(geometry=line, items=levee.items, height=height)
        heights.extend(list(height))
    average = np.nanmean(heights)
    output_merged.add(geometry=levee.geometry, items=levee.items, height=average)

output.write("levee_lines_example2.shp")
output_merged.write("levee_lines_example_merged2.shp")



# -*- coding: utf-8 -*-
"""
Created on Tue Feb 23 14:14:17 2021

@author: chris.kerklaan
"""
# First-party import
import math
import logging
from copy import deepcopy
from operator import attrgetter
from collections import namedtuple

# Third-party imports
import numpy as np
from osgeo import ogr
from itertools import chain

# Local imports
from .geometry import Geometry
from .point import Point, MultiPoint

# Globals
LINESTRING_COVERAGE = [
    ogr.wkbLineString,
    ogr.wkbLineString25D,
    ogr.wkbLineStringM,
    ogr.wkbLineStringZM,
]
MULTILINESTRING_COVERAGE = [
    ogr.wkbMultiLineString,
    ogr.wkbMultiLineString25D,
    ogr.wkbMultiLineStringM,
    ogr.wkbMultiLineStringZM,
]


# Logger
logger = logging.getLogger(__name__)


class LineString(Geometry):

    ogr_coverage = LINESTRING_COVERAGE

    def __init__(self, geometry: ogr.wkbLineString = None, epsg=None):
        super().__init__(geometry)
        self.check_type(LineString.ogr_coverage)
        if epsg:
            self.spatial_reference = epsg

    @classmethod
    def from_points(cls, points, flatten=True, epsg=None):
        """takes list of tuple points and creates an ogr linestring"""
        output_geom = ogr.Geometry(ogr.wkbLineString)
        for point in points:
            output_geom.AddPoint(*point)
        if flatten:
            output_geom.FlattenTo2D()

        return cls(output_geom, epsg)

    @classmethod
    def from_wkt(cls, wkt, flatten=True, epsg=None):
        """takes tuple points and creates an ogr point"""
        return cls(ogr.CreateGeometryFromWkt(wkt), epsg, flatten)

    def __iter__(self):
        """iterates over the vertices"""
        points = self.points
        for pid in range(0, len(points) - 1):
            yield LineString.from_points([points[pid], points[pid + 1]])

    @property
    def points_geometry(self):
        return [Point.from_point(point) for point in self.points]

    @property
    def start_point(self):
        return self.points_geometry[0]

    @property
    def end_point(self):
        return self.points_geometry[-1]

    @property
    def middle_point(self, interval=1):
        points = self.points_on_line(interval, vertices=True)
        return points[int(len(points) / 2)]

    def reversed(self):
        """reverses the line"""
        points = self.points
        points.reverse()
        return LineString.from_points(points)

    def intersection(self, geometry):
        """intersection of a linestring, returns a point, multipoint or none"""
        if self.Intersects(geometry):
            intersection = self.Intersection(geometry)

            if intersection.GetGeometryType() == ogr.wkbPoint:
                return Point(intersection)
            elif intersection.GetGeometryType() == ogr.wkbLineString:
                return LineString(intersection)
            else:
                return MultiPoint(intersection)

        else:
            logger.debug("Found no intersection")
            return None

    def points_on_line(
        self,
        interval=1,
        custom_interval=None,
        start=True,
        end=True,
        vertices=False,
        geometry=True,
    ):
        """Return points on a linestring

        Params:
            interval: distance between poitns
            custom_interval: a list of distances
            start: include start point
            end: include end point
            vertices: include vertices
            geometry: returns as a point
        """

        return points_on_line(
            self.points,
            interval,
            custom_interval,
            start,
            end,
            vertices,
            geometry,
        )

    def perpendicular_lines(
        self,
        distance,
        perpendicular_length,
        start=True,
        end=True,
        vertices=False,
    ):

        """returns a perpendicular linestring on point1 with a certain dist
        params:
            distance: distance between perpendicular lines
            perpendicular_length: length of the lines
        """
        return perpendicular_lines(
            self,
            distance,
            perpendicular_length,
            start=start,
            end=end,
            vertices=vertices,
        )

    def add_vertice(self, geometry: Point):
        """
        Returns a linestring with an extra vertice on the point
        If the vertice is already present, we will not adjust it.

        most likely there are two types of linestrings
        1. With one coordinate
        2. With a connection between coordinates
        this add vertice is for assumption 1.

        """
        # First we find all intersects
        distances = [(i, v.Distance(geometry)) for i, v in enumerate(self)]

        distances.sort(key=lambda tup: tup[-1])
        closest = distances[0]

        points = []
        snap_count = 0
        for i, vertice in enumerate(self):
            vertice_point = vertice.points[0]
            if vertice_point not in points:
                points.append(vertice_point)

            if i == closest[0]:
                points.append(geometry.point)
                snap_count += 1

        # last one
        if vertice.points[-1] not in points:
            points.append(vertice.points[-1])

        if snap_count == 0:
            logger.debug("snap not found")

        return LineString.from_points(points)

    def split(self, interval=1, custom_interval=None):
        points = self.points_on_line(interval, custom_interval, geometry=False)
        return [
            LineString.from_points([points[pid], points[pid + 1]])
            for pid in range(len(points) - 1)
        ]

    def split_on_vertices(self):
        return [vertice for vertice in self]

    def split_on_point(self, geometry: Point):

        points = self.points
        assert geometry.point in points

        index = points.index(geometry.point)
        line1 = LineString.from_points(points[: index + 1])
        line2 = LineString.from_points(points[index:])
        return [line1, line2]

    def as_multi(self):
        return MultiLineString.from_points(self.points)

    def transform(self, epsg):
        return LineString(self.reproject(epsg))

    def angle(self):
        return angle(*self.points)


class MultiLineString(Geometry):
    """Can also use linestrings"""

    ogr_coverage = MULTILINESTRING_COVERAGE

    def __init__(self, geometry: ogr.wkbMultiLineString = None, epsg=None):
        super().__init__(geometry)
        if epsg:
            self.spatial_reference = epsg

        self.check_type(MultiLineString.ogr_coverage)

    @classmethod
    def from_points(cls, points, flatten=True, epsg=None):
        """creates a ogr multilinestring from points"""

        output_geom = ogr.Geometry(ogr.wkbMultiLineString)

        # check if input is a single linestring
        if type(points[0]) == tuple:
            points = [[point] for point in points]

        for point in points:
            line = ogr.Geometry(ogr.wkbLineString)
            for p in point:
                line.AddPoint(*p)
            output_geom.AddGeometry(line)

        if flatten:
            output_geom.FlattenTo2D()

        return cls(output_geom, epsg)

    @classmethod
    def from_wkt(cls, wkt, flatten=True, epsg=None):
        """takes tuple points and creates an ogr point"""
        return cls(ogr.CreateGeometryFromWkt(wkt), epsg, flatten)

    def dissolve(self):
        return self.clean(False).to_single()[0]

    @property
    def points(self):  # overwrits the property of geometry class
        return [l.GetPoints() for l in self]

    def clean(self, intersection):
        """
        Returns a rearranged multilinestrings

        params:
            intersection: If True, splits the linestring on large intersects

        A multiline can be a multiline even if there are double
        coordinates and they are not arranged.
        This function arranges the multilinestring and cleans out all
        double points.

        """
        return MultiLineString(clean_multilinestring(self, intersection))

    def points_on_line(
        self,
        interval=1,
        custom_interval=None,
        start=True,
        end=True,
        vertices=False,
        geometry=True,
        single=True,
    ):

        all_points = []
        for points in self.points:
            all_points.append(
                points_on_line(
                    points,
                    interval,
                    custom_interval,
                    start,
                    end,
                    vertices,
                    geometry,
                )
            )

        if single:
            return list(chain.from_iterable(all_points))
        else:
            return all_points

    def perpendicular_lines(
        self,
        distance,
        perpendicular_length,
        start=True,
        end=True,
        vertices=False,
        single=True,
    ):

        """returns a perpendicular linestring on point1 with a certain dist
        params:
            distance: distance between perpendicular lines
            perpendicular_length: length of the lines
        """
        lines = []
        for linestring in self:
            lines.extend(
                perpendicular_lines(
                    self,
                    distance,
                    perpendicular_length,
                    start=start,
                    end=end,
                    vertices=vertices,
                )
            )

        if single:
            return list(chain.from_iterable(lines))
        else:
            return lines


def calc_dist(x1, y1, x2, y2):
    """returns the distance beteen two points"""
    return math.sqrt((x2 - x1) ** 2 + (y2 - y1) ** 2)


def lower_bound(x, l):
    if l[0] > x and not l[0] == x:
        return
    for i, y in enumerate(l):
        if y > x:
            return l[i - 1]


def points_with_distance(point1, point2, distance):
    """
    Returns a point on a certain distance between two points
    Note that the distance is the distance from the first point in the coordinate unit
    """
    t = distance / calc_dist(*point1, *point2)
    p = (
        ((1 - t) * point1[0] + t * point2[0]),
        ((1 - t) * point1[1] + t * point2[1]),
    )

    if type(p[0]) is np.ndarray:
        return t, (p[0][0], p[1][0])
    else:
        return t, p


def points_on_line(
    points,
    interval=1,
    custom_interval: list = None,
    start=True,
    end=True,
    vertices=False,
    geometry=True,
):
    """
    returns a point on the line for every interval that is given
    use combine to also include begin, endpoint and vertices
    Params:
        interval: distance between poitns
        custom_interval: a list of distances
        start: include start point
        end: include end point
        vertices: include vertices
        geometry: returns as a point
    """

    sections = list(
        np.cumsum(
            [
                calc_dist(*points[pid], *points[pid + 1])
                for pid in range(0, len(points) - 1)
            ]
        )
    )
    total_dist = sections[-1]

    if custom_interval is not None:
        new_sections = custom_interval
    else:
        new_sections = np.linspace(interval, total_dist, int(total_dist / interval))

    if vertices:
        new_sections = sorted(list(new_sections) + sections)

    new_points = []
    for i in new_sections:
        bound = lower_bound(i, sections)
        if not bound:
            index = 0
            dist = i
        else:
            index = sections.index(bound) + 1
            dist = i - sections[sections.index(bound)]

        ratio, point = points_with_distance(
            points[index], points[index + 1], distance=dist
        )

        if 0 <= ratio <= 1:
            new_points.append(point)
        else:
            pass

    if start:
        new_points.insert(0, points[0])

    if end:
        new_points.append(points[-1])

    if geometry:
        return [Point.from_point(point) for point in new_points]

    return new_points


def angle(pt1, pt2):
    x_diff = pt2[0] - pt1[0]
    y_diff = pt2[1] - pt1[1]
    return math.degrees(math.atan2(y_diff, x_diff))


def perpendicular_points(pt, bearing, dist):
    """returns perpendicular points at point"""
    bearing_pt1 = math.radians(bearing + 90)
    bearing_pt2 = math.radians(bearing - 90)
    points = []
    for bearing in [bearing_pt1, bearing_pt2]:
        x = pt[0] + dist * math.cos(bearing)
        y = pt[1] + dist * math.sin(bearing)
        points.append((x, y))
    return points


def perpendicular_line(pt1, pt2, dist):
    return LineString.from_points(perpendicular_points(pt1, angle(pt1, pt2), dist))


def perpendicular_lines(
    linestring,
    line_dist=10,
    perp_dist=10,
    start=True,
    end=True,
    vertices=False,
):
    """returns perpendicular lines on a linestring"""
    points = linestring.points_on_line(
        interval=line_dist,
        custom_interval=None,
        start=start,
        end=end,
        vertices=vertices,
        geometry=False,
    )

    lines = [
        perpendicular_line(points[index], points[index + 1], perp_dist)
        for index in range(0, len(points) - 1)
    ]

    # add last point
    if len(points) > 1:
        lines.append(perpendicular_line(points[-1], points[-2], perp_dist))

    return lines


def arrange_multiline(points):
    """arranges the points in a multilinestring
    e.g.,
    [[2, 1] ,[4, 5], [0, 1], [5, 6], [9, 8]]

    to
    [[0, 1], [1, 2], [4, 5], [5, 6], [8, 9]]

    """

    def distance(coords1, coords2):
        return math.sqrt(
            math.pow(coords1[0] - coords2[0], 2) + math.pow(coords1[1] - coords2[1], 2)
        )

    MinDistance = namedtuple("MinDistance", "target distance offset reverse_target")
    lines = deepcopy(points)
    sorted_multistring = [lines.pop(0)]

    while lines:
        min_distances = []

        for line in lines:
            source_a = sorted_multistring[0][0]
            source_b = sorted_multistring[-1][-1]

            target_a = line[0]
            target_b = line[-1]

            distances = [
                MinDistance(
                    target=line,
                    distance=distance(source_b, target_a),
                    offset=1,
                    reverse_target=False,
                ),
                MinDistance(
                    target=line,
                    distance=distance(source_a, target_a),
                    offset=-1,
                    reverse_target=True,
                ),
                MinDistance(
                    target=line,
                    distance=distance(source_b, target_b),
                    offset=1,
                    reverse_target=True,
                ),
                MinDistance(
                    target=line,
                    distance=distance(source_a, target_b),
                    offset=-1,
                    reverse_target=False,
                ),
            ]

            min_distance = min(distances, key=attrgetter("distance"))
            min_distances.append(min_distance)

        min_distance = min(min_distances, key=attrgetter("distance"))
        target = min_distance.target

        if min_distance.reverse_target:
            target.reverse()

        if min_distance.offset == 1:
            sorted_multistring.append(target)
        else:
            sorted_multistring.insert(0, target)

        lines.remove(target)

    return sorted_multistring


def clean_multilinestring(geometry, intersections=True):
    """cleans a messy linestring order and turns them into logical
    seperations.

    returns a list of linestrings

    intersections:
        Splits the lines at the intersctions of multiple lines

    """

    points = MultiLineString(geometry).points
    if len(points) == 1:
        return geometry

    arranged = arrange_multiline(points)

    # combine the the points of the arranged points,
    lines = []
    line = [arranged[0][0], arranged[0][1]]
    for i in range(len(arranged) - 1):
        coord2 = arranged[i][1]
        coord3, coord4 = arranged[i + 1][0], arranged[i + 1][1]

        match = coord2 == coord3
        if match:  # we have a match
            line.append(coord4)

        final_lap = i == len(arranged) - 2
        if not match:
            lines.append(line)
            line = [coord3, coord4]
        elif match and final_lap:
            lines.append(line)

    lines = [LineString.from_points(l) for l in lines]

    if intersections:
        # get intersects    intersects = []
        line_geometries = [(i, l) for i, l in enumerate(lines)]
        final_geometries = []
        for i1, line1 in line_geometries:

            new_lines = [line1]
            for i2, line2 in line_geometries:
                if i1 == i2:
                    continue

                # find intersections of the whole line
                intersects = []
                for new_line in new_lines:

                    if new_line.GetPointCount() < 2:
                        continue

                    intersect = new_line.intersection(line2)

                    if not intersect:
                        continue

                    points = []
                    if intersect.is_multi:
                        points = [p for p in intersect.points_geometry]
                    else:
                        points = [intersect]

                    for point in points:
                        on_start_point = point == new_line.start_point.point
                        on_end_point = point == new_line.end_point.point

                        if on_start_point or on_end_point:
                            continue

                        intersects.append(point)

                # split the line per point and add to the main line list
                line_count = len(new_lines)
                while len(intersects) > 0:
                    point = intersects.pop(0)

                    for i in range(line_count):
                        new_line = new_lines[i]
                        if len(new_line.GetPoints()) <= 2:
                            continue

                        if point.Intersects(new_line):
                            for geometry in new_line.split_on_point(point):
                                new_lines.append(geometry)

            if len(new_lines) > 1:
                del new_lines[0]

            for line in new_lines:
                final_geometries.append(line)
        lines = final_geometries

    multi = ogr.Geometry(5)
    for line in lines:
        multi.AddGeometry(line)

    return multi



# -*- coding: utf-8 -*-
"""
Created on Tue Feb 16 14:51:40 2021

@author: chris.kerklaan
"""
import math


class GeoTransform(tuple):
    def shifted(self, geometry, inflate=False):
        """
        Return shifted geo transform.

        :param geometry: geometry to match
        :param inflate: inflate to nearest top-left grid intersection.
        """
        values = list(self)
        index = self.get_indices(geometry, inflate=inflate)[1::-1]
        values[0], values[3] = self.get_coordinates(index)
        return self.__class__(values)

    def scaled(self, w, h):
        """
        Return shifted geo transform.

        :param f: scale the cellsize by this factor
        """
        p, a, b, q, c, d = self
        return self.__class__([p, a * w, b * h, q, c * w, d * h])

    def get_coordinates(self, indices):
        """Return x, y coordinates.

        :param indices: i, j tuple of integers or arrays.

        i corresponds to the y direction in a non-skew grid.
        """
        p, a, b, q, c, d = self
        i, j = indices
        return p + a * j + b * i, q + c * j + d * i

    def get_indices(self, geometry, inflate=False):
        """
        Return array indices tuple for geometry.

        :param geometry: geometry to subselect
        :param inflate: inflate envelope to grid, to make sure that
            the entire geometry is contained in resulting indices.
        """
        # spatial coordinates
        x1, x2, y1, y2 = geometry.GetEnvelope()

        # inverse transformation
        p, a, b, q, c, d = self
        e, f, g, h = get_inverse(a, b, c, d)

        # apply to envelope corners
        f_lo, f_hi = (math.floor, math.ceil) if inflate else (round, round)

        X1 = int(f_lo(e * (x1 - p) + f * (y2 - q)))
        Y1 = int(f_lo(g * (x1 - p) + h * (y2 - q)))
        X2 = int(f_hi(e * (x2 - p) + f * (y1 - q)))
        Y2 = int(f_hi(g * (x2 - p) + h * (y1 - q)))

        # prevent zero dimensions in case of inflate
        if inflate:
            if X1 == X2:
                X2 += 1
            if Y1 == Y2:
                Y1 -= 1

        return X1, Y1, X2, Y2

    def get_slices(self, geometry):
        """
        Return array slices tuple for geometry.

        :param geometry: geometry to subselect
        """
        x1, y1, x2, y2 = self.get_indices(geometry)
        return slice(y1, y2), slice(x1, x2)

    def get_window(self, geometry):
        """
        Return window dictionary for a geometry.

        :param geometry: geometry to subselect
        """
        x1, y1, x2, y2 = self.get_indices(geometry)
        return {"xoff": x1, "yoff": y1, "xsize": x2 - x1, "ysize": y2 - y1}


def get_inverse(a, b, c, d):
    """Return inverse for a 2 x 2 matrix with elements (a, b), (c, d)."""
    D = 1 / (a * d - b * c)
    return d * D, -b * D, -c * D, a * D
